//===- RawMemProfReader.h - Instrumented memory profiling reader *- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
//
// This file just includes MemProfReader.h for compatibility with
// out-of-tree users.
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_PROFILEDATA_RAWMEMPROFREADER_H_
#define LLVM_PROFILEDATA_RAWMEMPROFREADER_H_

#include "llvm/ProfileData/MemProfReader.h"

#endif // LLVM_PROFILEDATA_RAWMEMPROFREADER_H_
